import numpy as np
import tensorflow as tf

class SamplerBase:
    """Other classes will be used this helper class."""

    def __init__(self, dtype):
        """Base class for sampling mesh data for training.

        Initializes instance variables for storing sampled data and solutions.
        """
        self.time_domain_sampled = None
        self.spatial_domain_sampled = None
        self.solution_sampled = None
        self.solution_names = None
        dtype = 'float32'
        self.np_dtype = np.dtype(dtype)

    def concatenate_solutions(self, flatten_mesh):
        """Concatenate dictionary of sampled solution data.

        :param flatten_mesh: Flattened mesh data.
        :return: Flattened mesh data with concatenated solutions.
        """
        flatten_mesh = list(flatten_mesh)
        concatenated_solutions = [
            flatten_mesh[2][solution_name] for solution_name in self.solution_names
        ]
        flatten_mesh[2] = np.concatenate(concatenated_solutions, axis=-1)

        return flatten_mesh

    def sample_mesh(self, num_sample, flatten_mesh):
        """Sample the mesh data for training.

        :param num_sample: Number of samples to generate.
        :param flatten_mesh: Flattened mesh data.
        :return: Sampled spatial, time, and solution data.
        """

        flatten_mesh = self.concatenate_solutions(flatten_mesh)

        if num_sample is None:
            return self.convert_to_tensor(flatten_mesh)
        else:
            idx = np.random.choice(range(flatten_mesh[0].shape[0]), num_sample, replace=False)
            return self.convert_to_tensor(
                (flatten_mesh[0][idx, :], flatten_mesh[1][idx, :], flatten_mesh[2][idx, :])
            )

    def convert_to_tensor(self, arrays):
        """Convert NumPy arrays to TensorFlow tensors.

        :param arrays: List of NumPy arrays to convert.
        :return: List of converted TensorFlow tensors.
        """

        return [tf.constant(array.astype(self.np_dtype)) for array in arrays]

    def loss_fn(self, inputs, loss, **functions):
        """Compute the loss function based on given inputs and functions.

        :param inputs: Input data for computing the loss.
        :param loss: Loss variable.
        :param functions: Additional functions required for loss computation.
        """

        pass

    def requires_grad(self, x, t, enable_grad=True):
        """Set the requires_grad attribute for tensors in the input list.

        :param x: List of tensors to modify requires_grad attribute.
        :param t: Tensor to modify requires_grad attribute.
        :param enable_grad: Boolean indicating whether to enable requires_grad or not.
        :return: Modified list of tensors and tensor.
        """
        if t is not None:
            t = t.requires_grad_(enable_grad)
        x = [x_.requires_grad_(enable_grad) for x_ in x]

        return x, t

    @property
    def mean(self):
        """Calculate the mean of the concatenated input data along each column.

        :return: A numpy array containing the mean values along each column.
        """

        x, t, _ = self[:]
        inputs = np.concatenate((*x, t), 1).astype(np.float32)

        return inputs.mean(0, keepdims=True)

    @property
    def std(self):
        """Calculate the standard deviation of the concatenated input data along each column.

        :return: A numpy array containing the standard deviation values along each column.
        """

        x, t, _ = self[:]
        inputs = np.concatenate((*x, t), 1).astype(np.float32)

        return inputs.std(0, keepdims=True)

    
    def shuffle(self):
        random_indices = tf.random.shuffle(tf.range(len(self)))
        self.spatial_domain_sampled = [tf.gather(spatial_domain, random_indices) \
                                       for spatial_domain in self.spatial_domain_sampled]
        if self.time_domain_sampled is not None:
            self.time_domain_sampled = tf.gather(self.time_domain_sampled, random_indices)
        if self.solution_sampled is not None:
            self.solution_sampled = [tf.gather(solution, random_indices) \
                                     for solution in self.solution_sampled]

        
    def __len__(self):
        """Get the number of sampled data points.

        :return: The number of sampled data points.
        """

        return len(self.time_domain_sampled)

    def __getitem__(self, idx):
        """Get a specific sampled data point using indexing. In some cases, we may not have
        `time_domain` and `solution_domain`. For example, in periodic boundary condition, there is
        not `solution_domain`.

        :param idx: Index of the desired data point.
        :return: Tuple containing spatial, time, and solution data for the indexed point.
        """

        spatial_domain = [spatial_domain[idx] for spatial_domain in self.spatial_domain_sampled]

        time_domain = None
        if self.time_domain_sampled is not None:
            time_domain = self.time_domain_sampled[idx]

        solution_domain = None
        if self.solution_sampled is not None:
            solution_domain = {
                solution_name: self.solution_sampled[i][idx]
                for i, solution_name in enumerate(self.solution_names)
            }

        return (spatial_domain, time_domain, solution_domain)




class PINNDataModule:
    def __init__(self,
                 train_datasets,
                 val_dataset,
                 test_dataset=None,
                 pred_dataset=None,
                 batch_size=None,
                ):
        """Initialize a `MNISTDataModule`.

        :param train_datasets: train datasets.
        :param val_dataset: validation dataset.
        """
        
        self.train_datasets = train_datasets
        self.val_datasets = val_dataset
        self.test_datasets = test_dataset
        self.pred_datasets = pred_dataset

        self.train_data = None
        self.val_data = None
        self.test_data = None
        self.pred_data = None
        self.batch_size = batch_size

        self.function_mapping = {}

    
    def setup(self, stage: str):
        """Load data. Set variables: `self.data_train`, `self.data_val`, `self.data_test`.

        This method is called by Lightning before `trainer.fit()`, `trainer.validate()`, `trainer.test()`, and
        `trainer.predict()`, so be careful not to execute things like random split twice! Also, it is called after
        `self.prepare_data()` and there is a barrier in between which ensures that all the processes proceed to
        `self.setup()` once the data is prepared and available for use.

        :param stage: The stage to setup. Either `"fit"`, `"validate"`, `"test"`, or `"predict"`. Defaults to ``None``.
        """
        if stage=='fit':
            if not self.train_data:
                self.train_data = {}
                self.set_mode_for_discrete_mesh()
    
                for train_dataset in self.train_datasets:
                    if self.batch_size is not None:
                        self.train_data[str(train_dataset.loss_fn)] = train_dataset
                    else:
                        self.train_data[str(train_dataset.loss_fn)] = train_dataset[:]
                    self.function_mapping[str(train_dataset.loss_fn)] = train_dataset.loss_fn
        
        if stage=='val':
            if not self.val_data:
                self.val_solution_names = []
                self.val_data = {}
                if not isinstance(self.val_datasets, list):
                    self.val_datasets = [self.val_datasets]
                for val_dataset in self.val_datasets:
                    if self.batch_size is not None:
                        self.val_data[str(val_dataset.loss_fn)] = val_dataset
                    else:
                        self.val_data[str(val_dataset.loss_fn)] = val_dataset[:]
                    self.function_mapping[str(val_dataset.loss_fn)] = val_dataset.loss_fn
                    self.val_solution_names.extend(val_dataset.solution_names)

        if stage=='test':
            if not self.test_data:
                self.test_data = {}
                if not isinstance(self.test_datasets, list):
                    self.test_datasets = [self.test_datasets]
                for test_dataset in self.test_datasets:
                    self.test_data[str(test_dataset.loss_fn)] = test_dataset[:]
                    self.function_mapping[str(test_dataset.loss_fn)] = test_dataset.loss_fn

        if stage=='pred':
            if not self.pred_data:
                self.pred_data = {}
                if not isinstance(self.pred_datasets, list):
                    self.pred_datasets = [self.pred_datasets]
                for pred_dataset in self.pred_datasets:
                    self.pred_data[str(pred_dataset.loss_fn)] = pred_dataset[:]
                    self.function_mapping[str(pred_dataset.loss_fn)] = pred_dataset.loss_fn
        
     

    def set_mode_for_discrete_mesh(self):
        """This function will figuere out which training datasets are for discrete.

        Then set the mode value that will be used for Runge–Kutta methods
        """

        mesh_idx = [
            (train_dataset.idx_t, train_dataset)
            for train_dataset in self.train_datasets
            if type(train_dataset).__name__ == "DiscreteMeshSampler"
        ]

        mesh_idx = sorted(mesh_idx, key=lambda x: x[0])

        if len(mesh_idx) == 1:
            mesh_idx[0][1].mode = "forward_discrete"
        elif len(mesh_idx) == 2:
            mesh_idx[0][1].mode = "inverse_discrete_1"
            mesh_idx[1][1].mode = "inverse_discrete_2"

        mesh_idx.clear()


        
    def train_dataloader(self):
        """Create and return the train dataloader.

        :return: The train dataloader.
        """
        return self.train_data

    def val_dataloader(self):
        """Create and return the validation dataloader.

        :return: The train dataloader.
        """
        return self.val_data
        
    def test_dataloader(self):
        """Create and return the test dataloader.

        :return: The train dataloader.
        """
        return self.test_data

    def predict_dataloader(self):
        """Create and return the predict dataloader.

        :return: The train dataloader.
        """
        return self.pred_data 


if __name__ == "__main__":
    _ = PINNDataModule(None, None)

        


from typing import List, Dict

import numpy as np
import tensorflow as tf

class FCN(object):
    def __init__(self, layers, lb, ub, output_names, discrete: bool = False, dtype='float32') -> None:
        """Initialize a `FCN` module.

        :param layers: The list indicating number of neurons in each layer.
        :param lb: Lower bound for the inputs.
        :param ub: Upper bound for the inputs.
        :param output_names: Names of outputs of net.
        :param discrete: If the problem is discrete or not.
        """
        super().__init__()
        
        if dtype == 'float32':
            self.dtype = tf.float32
        else:
            self.dtype = None

        self.lb = tf.constant(lb, tf.float32)
        self.ub = tf.constant(ub, tf.float32)

        self.model = self.initalize_net(layers)
        self.output_names = output_names
        self.discrete = discrete
    
    def create_layer(self, input, num_output, initializer, activation=None, dtype=None):
        if activation is None:
            return tf.keras.layers.Dense(
                        num_output,
                        kernel_initializer=initializer,
                        dtype=tf.float32,
                    )(input)
        
        if dtype is not None:
            return tf.keras.layers.Dense(
                        num_output,
                        activation="tanh",
                        kernel_initializer=initializer,
                        dtype=dtype,
                    )(input)
        else:
            return tf.keras.layers.Dense(
                        num_output,
                        activation="tanh",
                        kernel_initializer=initializer,
                    )(input)
            
    def initalize_net(self, layers):
        initializer = tf.initializers.GlorotUniform(seed=1234)
        
        if self.dtype is None:
            inputs = tf.keras.layers.Input(shape=(layers[0],))
        else:
            inputs = tf.keras.layers.Input(shape=(layers[0],), dtype=self.dtype)
        
        z = self.create_layer(inputs, layers[1], initializer, "tanh", self.dtype)
        
        for i in range(1, len(layers) - 2):
            z = self.create_layer(z, layers[i + 1], initializer, "tanh", self.dtype)

        outputs = self.create_layer(z, layers[-1], initializer)

        neural_net = tf.keras.Model(inputs=inputs, outputs=outputs)
        neural_net.summary()
        
        return neural_net
        
    def __call__(self, spatial: List[tf.Tensor], time: tf.Tensor) -> Dict[str, tf.Tensor]:
        """Perform a single forward pass through the network.

        :param spatial: List of input spatial tensors.
        :param time: Input tensor representing time.
        :return: A tensor of solutions.
        """

        # Discrete Mode
        if self.discrete:
            if len(spatial) == 2:
                x, y = spatial
                z = tf.concat((x, y), 1)
            elif len(spatial) == 3:
                x, y, z = spatial
                z = tf.concat((x, y, z), 1)
            else:
                z = spatial[0]
            z = 2.0 * (z - self.lb[:-1]) / (self.ub[:-1] - self.lb[:-1]) - 1.0

        # Continuous Mode
        else:
            if len(spatial) == 1:
                x = spatial[0]
                z = tf.concat((x, time), 1)
            elif len(spatial) == 2:
                x, y = spatial
                z = tf.concat((x, y, time), 1)
            else:
                x, y, z = spatial
                z = tf.concat((x, y, z, time), 1)
            z = 2.0 * (z - self.lb) / (self.ub - self.lb) - 1.0

        z = self.model(z)

        # Discrete Mode
        if self.discrete:
            outputs_dict = {name: z for i, name in enumerate(self.output_names)}

        # Continuous Mode
        else:
            outputs_dict = {name: z[:, i : i + 1] for i, name in enumerate(self.output_names)}
            
        return outputs_dict

class NetHFM(object):
    """A simple fully-connected neural net for solving equations.

    In this model, mean and std will be used for normalization of input data. Also, weight
    normalization will be done.
    """
    output_names: List[str]
    
    def __init__(self, mean, std, layers: List, output_names: List):
        super().__init__()
        """Initialize a `NetHFM` module.

        :param mesh: The number of layers.
        :param layers: The list indicating number of neurons in each layer.
        :param output_names: Names of outputs of net.
        """
        self.num_layers = len(layers)
        self.output_names = output_names
        self.trainable_variables = []

        self.X_mean = tf.constant(mean, dtype=tf.float32)
        self.X_std = tf.constant(std, dtype=tf.float32)

        self.initalize_net(layers)
        
    def initalize_net(self, layers: List) -> None:
        """Initialize the neural network weights, biases, and gammas.

        :param layers: The list indicating number of neurons in each layer.
        """

        self.weights = []
        self.biases = []
        self.gammas = []
        
        for l in range(0,self.num_layers-1):
            in_dim = layers[l]
            out_dim = layers[l+1]
            W = np.random.normal(size=[in_dim, out_dim])
            b = np.zeros([1, out_dim])
            g = np.ones([1, out_dim])
            # tensorflow variables
            self.weights.append(tf.Variable(W, dtype=tf.float32, trainable=True))
            self.biases.append(tf.Variable(b, dtype=tf.float32, trainable=True))
            self.gammas.append(tf.Variable(g, dtype=tf.float32, trainable=True))

        self.trainable_variables.extend(self.weights)
        self.trainable_variables.extend(self.biases)
        self.trainable_variables.extend(self.gammas)
    
    def __call__(self, spatial: List[tf.Tensor], time: tf.Tensor) -> Dict[str, tf.Tensor]:
        """Perform a forward pass through the network.

        :param spatial: List of input spatial tensors.
        :param time: Input tensor representing time.
        :return: A dictionary with output names as keys and corresponding output tensors as values.
        """
        if len(spatial) == 1:
            x = spatial[0]
            H = tf.concat((x, time), 1)
        elif len(spatial) == 2:
            x, y = spatial
            H = tf.concat((x, y, time), 1)
        else:
            x, y, z = spatial
            H = tf.concat((x, y, z, time), 1)
        
        H = (H - self.X_mean) / self.X_std

        for i, (W, b, g) in enumerate(zip(self.weights, self.biases, self.gammas)):
            # weight normalization
            V = W/tf.norm(W, axis = 0, keepdims=True)
            # matrix multiplication
            H = tf.matmul(H, V)
            # add bias
            H = g * H + b
            # activation
            if i < self.num_layers - 2:
                H = H * tf.sigmoid(H)
        
        outputs_dict = {name: H[:, i : i + 1] for i, name in enumerate(self.output_names)}

        return outputs_dict


if __name__ == "__main__":
    _ = FCN()
    _ = NetHFM()


from typing import List, Dict, Callable, Any, Tuple, Union

import tensorflow as tf
import sys, os, logging, time

from pinnstf2.utils import fwd_gradient, gradient
from pinnstf2.utils import (
    fix_extra_variables,
    mse,
    relative_l2_error,
    sse
)

class PINNModule:
    def __init__(
        self,
        net,
        pde_fn: Callable[[Any, ...], tf.Tensor],
        optimizer: tf.keras.optimizers.Adam = tf.keras.optimizers.Adam,
        loss_fn: str = "sse",
        extra_variables: Dict[str, Any] = None,
        output_fn: Callable[[Any, ...], tf.Tensor] = None,
        runge_kutta=None,
        jit_compile: bool = True,
        amp: bool = False,
        dtype: str = 'float32'
    ) -> None:
        """
        Initialize a `PINNModule`.

        :param net: The neural network model to be used for approximating solutions.
        :param pde_fn: The partial differential equation (PDE) function defining the PDE to solve.
        :param optimizer: The optimizer used for training the neural network.
        :param loss_fn: The name of the loss function to be used. Default is 'sse' (sum of squared errors).
        :param extra_variables: Additional variables used in the model, provided as a dictionary. Default is None.
        :param output_fn: A function applied to the output of the network, for post-processing or transformations.
        :param runge_kutta: An optional Runge-Kutta method implementation for solving discrete problems. Default is None.
        :param jit_compile: If True, TensorFlow's JIT compiler will be used for optimizing computations. Default is True.
        :param amp: Automatic mixed precision (amp) for optimizing training performance. Default is False.
        :param dtype: Data type to be used for the computations. Default is 'float32'.
        """
        super().__init__()
        
        self.net = net
        self.tf_dtype = tf.as_dtype(dtype)
        
        if hasattr(self.net, 'model'):
            self.trainable_variables = self.net.model.trainable_variables
        else:
            self.trainable_variables = self.net.trainable_variables
        (self.trainable_variables,
         self.extra_variables) = fix_extra_variables(self.trainable_variables, extra_variables, self.tf_dtype)
        self.output_fn = output_fn
        self.rk = runge_kutta
        self.pde_fn = pde_fn
        self.opt = optimizer()
        self.amp = amp
        if self.amp:
            self.opt = tf.keras.mixed_precision.LossScaleOptimizer(self.opt)

        if jit_compile:
            self.train_step = tf.function(self.train_step, jit_compile=True)
            self.eval_step = tf.function(self.eval_step, jit_compile=True)
        else:
            self.train_step = tf.function(self.train_step, jit_compile=False)
            self.eval_step = tf.function(self.eval_step, jit_compile=False)
        
        if loss_fn == "sse":
            self.loss_fn = sse
        elif loss_fn == "mse":
            self.loss_fn = mse
        
        self.time_list = []
        self.functions = {
            "runge_kutta": self.rk,
            "forward": self.forward,
            "pde_fn": self.pde_fn,
            "output_fn": self.output_fn,
            "extra_variables": self.extra_variables,
            "loss_fn": self.loss_fn,
        }

    def forward(self, spatial: List[tf.Tensor], time: tf.Tensor) -> Dict[str, tf.Tensor]:
        """Perform a forward pass through the model `self.net`.

        :param spatial: List of input spatial tensors.
        :param time: Input tensor representing time.
        :return: A tensor of solutions.
        """
        outputs = self.net(spatial, time)
        if self.output_fn:
            outputs = self.output_fn(outputs, *spatial, time)
        return outputs
    
    def model_step(
        self,
        batch: Dict[
            str,
            Union[
                Tuple[tf.Tensor, tf.Tensor, tf.Tensor], Tuple[tf.Tensor, tf.Tensor]
            ],
        ],
    ) -> Tuple[tf.Tensor, Dict[str, tf.Tensor]]:
        """Perform a single model step on a batch of data.

        :param batch: A batch of data (a tuple) containing the
        input tensor of different conditions and data.

        :return: A tuple containing (in order):
            - A tensor of losses.
            - A dictionary of predictions.
        """
        loss = 0.0
        for loss_fn_name, data in batch.items():
            loss, preds = self.function_mapping[loss_fn_name](data, loss, self.functions)

        return loss, preds

    def train_step(self, batch):
        """
        Performs a single training step, including forward and backward passes.
    
        :param batch: The input batch of data for training.
        :return: The calculated loss and any extra variables to be used outside.
        """

        # Use GradientTape for automatic differentiation - to record operations for the forward pass
        with tf.GradientTape() as tape:
            loss, pred = self.model_step(batch)

            # If automatic mixed precision (amp) is enabled, scale the loss to prevent underflow
            if self.amp:
                scaled_loss = self.opt.get_scaled_loss(loss)

        # If amp is enabled, compute gradients w.r.t the scaled loss
        if self.amp:
            scaled_grad = tape.gradient(scaled_loss, self.trainable_variables)
            gradients = self.opt.get_unscaled_gradients(scaled_grad)
        else:
            gradients = tape.gradient(loss, self.trainable_variables)

        # Apply the calculated gradients to the model's trainable parameters
        self.opt.apply_gradients(zip(gradients, self.trainable_variables))   
        
        return loss, self.extra_variables

    def eval_step(
        self, batch
    ) -> Tuple[tf.Tensor, Dict[str, tf.Tensor], Dict[str, tf.Tensor]]:
        """Perform a single evaluation step on a batch of data.

        :param batch: A batch of data containing input tensors and conditions.
        :return: A tuple containing loss, error dictionary, and predictions.
        """

        x, t, u = list(batch.values())[0]
                
        loss, preds = self.model_step(batch)

        if self.rk:
            error_dict = {
                solution_name: relative_l2_error(
                    preds[solution_name][:, -1][:, None], u[solution_name]
                )
                for solution_name in self.val_solution_names
            }

        else:
            error_dict = {
                solution_name: relative_l2_error(preds[solution_name], u[solution_name])
                for solution_name in self.val_solution_names
            }
                
        return loss, error_dict, preds
   
    def validation_step(self, batch):
        """Perform a single validation step on a batch of data from the validation set.

        :param batch: A batch of data (a dict).
        :param batch_idx: The index of the current batch.
        """
        
        loss, error_dict, _ = self.eval_step(batch)
        
        return loss, error_dict    
    
    def test_step(self, batch):
        """Perform a single predict step on a batch of data from the test set.

        :param batch: A batch of data (a dict).
        :param batch_idx: The index of the current batch.
        """

        loss, error, _ = self.eval_step(batch)

        return loss, error
                
    def predict_step(self, batch):
        """Perform a single predict step on a batch of data from the prediction set.

        :param batch: A batch of data (a dict).
        :param batch_idx: The index of the current batch.
        """

        _, _, preds = self.eval_step(batch)

        return preds
                


import logging

def get_pylogger(name: str = __name__) -> logging.Logger:
    """Initializes a TensorFlow 2 compatible python command line logger.

    :param name: The name of the logger, defaults to ``__name__``.

    :return: A logger object.
    """
    logger = logging.getLogger(name)

    return logger



import tensorflow as tf

from typing import Dict, List, Union, Optional, Tuple


def sse(loss: tf.Tensor,
        preds: Dict[str, tf.Tensor],
        target: Union[Dict[str, tf.Tensor], None] = None,
        keys: Union[List[str], None] = None,
        mid: Union[int, None] = None) -> tf.Tensor:
    """Calculate the sum of squared errors (SSE) loss for given predictions and optional targets.

    :param loss: Loss variable.
    :param preds: Dictionary containing prediction tensors for different keys.
    :param target: Dictionary containing target tensors (optional).
    :param keys: List of keys for which to calculate SSE loss (optional).
    :param mid: Index to separate predictions for mid-point calculation (optional).
    :return: Calculated SSE loss.
    """
    
    if keys is None:
        return loss

    for key in keys:
        if target is None and mid is None:
            loss = loss + tf.reduce_sum(tf.square(preds[key]))
        elif target is None and mid is not None:
            loss = loss + tf.reduce_sum(tf.square(preds[key][:mid] - preds[key][mid:]))
        elif target is not None:
            loss = loss + tf.reduce_sum(tf.square(preds[key] - target[key]))

    return loss

def mse(loss: tf.Tensor,
        preds: Dict[str, tf.Tensor],
        target: Union[Dict[str, tf.Tensor], None] = None,
        keys: Union[List[str], None] = None,
        mid: Union[int, None] = None) -> tf.Tensor:
    """Calculate the mean squared error (MSE) loss for given predictions and optional targets.

    :param loss: Loss variable.
    :param preds: Dictionary containing prediction tensors for different keys.
    :param target: Dictionary containing target tensors (optional).
    :param keys: List of keys for which to calculate SSE loss (optional).
    :param mid: Index to separate predictions for mid-point calculation (optional).
    :return: Calculated MSE loss.
    """
    
    if keys is None:
        return loss

    for key in keys:
        if target is None:
            loss = loss + tf.reduce_mean(tf.square(preds[key]))
        elif target is None and mid is not None:
            loss = loss + tf.reduce_mean(tf.square(preds[key][:mid] - preds[key][mid:]))
        elif target is not None:
            loss = loss + tf.reduce_mean(tf.square(preds[key] - target[key]))

    return loss


def relative_l2_error(preds, target):
    """Calculate the relative L2 error between predictions and target tensors.

    :param preds: Predicted tensors.
    :param target: Target tensors.
    :return: Relative L2 error value.
    """
    
    #return tf.sqrt(tf.reduce_mean(tf.square(preds - target))/tf.reduce_mean(tf.square(target)))
    return tf.sqrt(tf.reduce_mean(tf.square(preds - target))/tf.reduce_mean(tf.square(target - tf.reduce_mean(target))))


def fix_extra_variables(trainable_variables, extra_variables, dtype):
    """Convert extra variables to tf tensors with gradient tracking. These variables are
    trainables in inverse problems.

    :param extra_variables: Dictionary of extra variables to be converted.
    :return: Dictionary of converted extra variables as tf tensors with gradients.
    """
    
    if extra_variables is None:
        return trainable_variables, None
    extra_variables_dict = {}
    for key in extra_variables:
        variable =  tf.Variable(extra_variables[key], dtype=tf.float32, trainable=True)
        extra_variables_dict[key] = variable
        trainable_variables.append(variable)
    return trainable_variables, extra_variables_dict

def fix_predictions(preds_dict):
    for sol_key, pred in preds_dict.items():
        preds_dict[sol_key] = pred.numpy()
    return preds_dict


from pinnstf2.utils.gradient import (
    gradient,
    fwd_gradient
)
from pinnstf2.utils.module_fn import (
    fix_extra_variables,
    sse,
    mse,
    relative_l2_error,
    fix_predictions,
)
from pinnstf2.utils.utils import (
    extras,
    get_metric_value,
    load_data,
    load_data_txt,
    task_wrapper,
    set_mode
)
from pinnstf2.utils.pylogger import get_pylogger
from pinnstf2.utils.plotting import (
    plot_ac,
    plot_burgers_continuous_forward,
    plot_burgers_continuous_inverse,
    plot_burgers_discrete_forward,
    plot_burgers_discrete_inverse,
    plot_kdv,
    plot_navier_stokes,
    plot_schrodinger,
)


import tensorflow as tf

def gradient(dy, dx, grad_ys=None):
    if grad_ys is None:
        dy_dx = tf.gradients(dy, dx)
    else:
        dy_dx = tf.gradients(dy, dx, grad_ys=grad_ys)
    if len(dy_dx)==1:
        dy_dx = dy_dx[0]
    return dy_dx

def fwd_gradient(dy, dx):
    dummy = tf.ones_like(dy)
    G = tf.gradients(dy, dx, grad_ys=dummy)[0]
    Y_x = tf.gradients(G, dummy)[0]
    return Y_x


import time

import numpy as np
import tensorflow as tf

from tqdm import tqdm
from pinnstf2 import utils

log = utils.get_pylogger(__name__)


class Trainer:
    """
    Trainer Class
    """
    
    def __init__(self,
                 max_epochs,
                 min_epochs: int=1,
                 enable_progress_bar: bool=True,
                 check_val_every_n_epoch: int = 1,
                 default_root_dir: str = ""):
        """
        Initialize the Trainer class with specified training parameters.

        :param max_epochs: Maximum number of training epochs.
        :param min_epochs: Minimum number of training epochs.
        :param enable_progress_bar: Flag to enable/disable the progress bar.
        :param check_val_every_n_epoch: Frequency of validation checks within epochs.
        :param default_root_dir: Default directory for saving model-related files.
        """
        
        self.min_epochs = min_epochs
        self.max_epochs = max_epochs
        self.enable_progress_bar = enable_progress_bar
        self.check_val_every_n_epoch = check_val_every_n_epoch
        self.callback_metrics = {}
        self.current_epoch = 0
        self.default_root_dir = default_root_dir
        self.time_list = []

    def callback_pbar(self, loss_name, loss, extra_variables=None):
        """
        Update and format the string for the tqdm progress bar.

        :param loss_name: Name of the loss function.
        :param loss: Loss value.
        :param extra_variables: Additional trainable variables to include in progress bar.
        :return: Formatted string with loss and extra variables values.
        """
        res = f"{loss_name}: {loss:.4f}"
        self.callback_metrics[loss_name] = loss#.numpy()
        
        if extra_variables:
            disc = []
            for key, value in extra_variables.items():
                disc.append(f"{key}: {value:.4f}")
                self.callback_metrics[key] = value.numpy()
                
            extra_info = ', '.join(disc)
            res = f"{res}, {extra_info}"
        
        return res

    def set_callback_metrics(self, loss_name, loss, extra_variables=None):
        """
        Set callback metrics such as loss and additional variables.

        :param loss_name: Name of the loss function.
        :param loss: Loss value.
        :param extra_variables: Additional trainable variables to be logged.
        """
        
        self.callback_metrics[loss_name] = loss#.numpy()
        
        if extra_variables:
            for key, value in extra_variables.items():
                self.callback_metrics[key] = value.numpy()
                            
    def initalize_tqdm(self, max_epochs, mode='training', leave=True):
        """
        Initialize and return a tqdm progress bar object.

        :param max_epochs: Maximum number of epochs for which the progress bar will run.
        :param mode: Mode can be `training` or `validation`.
        :param leave: Ensures that once the bar is closed, it doesn't occupy space in the console.
        :return: Initialized tqdm progress bar object.
        """
        
        return tqdm(
                total = max_epochs,
                bar_format = mode +": {n_fmt}/{total_fmt} {percentage:3.0f}%|{bar}|[{elapsed}<{remaining}, "
                "{rate_fmt}{postfix}, "
                "{desc}]",
                leave=leave,
                )

    def next_batch(self, dataloader, current_index, dataset_size, shuffle=True):
        """
        This function will return the next batch of data and current index.
        
        :param dataloader: The dictionary of dataloaders.
        :param current_index: The list indicating the current index.
        :param dataset_size: The list indicating the size of each dataset.
        :param shuffle: The flag for shuffling the data.

        :return: A new dictionary containing batch data and a list showing current index.
        """
        
        batch_data = {}
        for i, (key, data) in enumerate(dataloader.items()):  
            if shuffle:
                batch_data[key] = data[current_index[i]:current_index[i] + self.batch_size]
                current_index[i] = current_index[i] + self.batch_size
                
                if current_index[i] + self.batch_size >= dataset_size[i]:
                        data.shuffle()
                        current_index[i] = 0 
            else:
                if current_index[i] + self.batch_size >= dataset_size[i]:
                    batch_data[key] = data[current_index[i]:dataset_size[i]]
                else:
                    batch_data[key] = data[current_index[i]:current_index[i] + self.batch_size]
                    current_index[i] = current_index[i] + self.batch_size
                    
        return batch_data, current_index
        
    def initalize_batch_tracker(self, dataloader, shuffle=True):
        """
        Initialize tracking for batch processing.
        
        :param dataloader: The dictionary of dataloaders.
        :param shuffle: The flag for shuffling the data.

        :return: A list showing current index and a list indicating the size of datasets.
        """
        
        current_index = []
        dataset_size = []

        for i, (key, data) in enumerate(dataloader.items()):  
            current_index.append(0)
            dataset_size.append(len(data))
            if shuffle:
                data.shuffle()

        return current_index, dataset_size

    def prepare_model(self, model, datamodule):
        """
        Fix function mappings.
        
        :param datamodule: :param datamodule: The data module providing the data.
        """

        # Store validation solution names and function mappings in the model
        model.val_solution_names = datamodule.val_solution_names
        model.function_mapping = datamodule.function_mapping

        return model
    
    def fit(self, model, datamodule):    
        """
        Main function to fit the model on the provided data.

        :param model: The PINNModule.
        :param datamodule: The data module providing the data.
        """

        # Prepare the data for training and validation
        datamodule.setup('fit')
        datamodule.setup('val')

        # Load training and validation data using dataloaders
        train_dataloader = datamodule.train_dataloader()
        val_dataloader = datamodule.val_dataloader()

        self.batch_size = datamodule.batch_size
        
        model = self.prepare_model(model, datamodule)

        
        # Set up the dataset for batch training
        if datamodule.batch_size is not None:
            self.train_current_index, self.train_dataset_size = self.initalize_batch_tracker(train_dataloader)
            self.eval_current_index, self.eval_dataset_size = self.initalize_batch_tracker(val_dataloader, shuffle=False)
        
        # Initialize the progress bar if enabled
        if self.enable_progress_bar:
            self.pbar = self.initalize_tqdm(self.max_epochs)
        
        for epoch in range(self.current_epoch, self.max_epochs):
            
            start_time=time.time()

            self.train_loop(model, train_dataloader)
            
            elapsed_time = time.time() - start_time
            self.time_list.append(elapsed_time)
        
            # Perform validation at specified intervals
            if epoch % self.check_val_every_n_epoch == 0:
                self.eval_loop(model, val_dataloader)

        if self.enable_progress_bar:
            self.pbar.close()

    
    def train_loop(self, model, train_dataloader):
        """
        Training loop. 

        :param model: The PINNModule.
        :param train_dataloader: The training data.
        """
        
        # Process the data in batches if batch size is specified
        if self.batch_size is not None:
            
            train_data, self.train_current_index = self.next_batch(train_dataloader,
                                                                   self.train_current_index,
                                                                   self.train_dataset_size)
            loss, extra_variables = model.train_step(train_data)
            
        else:
            # If no batching is used, pass the entire dataloader to the train_step
            loss, extra_variables = model.train_step(train_dataloader)

        self.set_callback_metrics('train/loss', loss.numpy(), extra_variables)
        
        if self.enable_progress_bar:
            self.pbar.update(1)
            description = self.callback_pbar('train/loss', loss.numpy(), extra_variables)
            self.pbar.set_description(description)
            self.pbar.refresh() 

    
    def eval_loop(self, model, eval_dataloader, mode='validation'):
        """
        Evaluation loop. 

        :param model: The PINNModule.
        :param eval_dataloader: The validation data.
        """

        
        
        if self.batch_size is not None:

            iter = (self.eval_dataset_size[0]//self.batch_size) + 1
            
            if self.enable_progress_bar:
                self.eval_pbar = self.initalize_tqdm(iter, mode=mode, leave=False)
                
            loss = []
            error_dict = {name: [] for name in model.val_solution_names}
            
            
            for i in range(iter):
                eval_data, self.eval_current_index = self.next_batch(eval_dataloader,
                                                                    self.eval_current_index,
                                                                    self.eval_dataset_size,
                                                                    shuffle=False)
                loss_i, error_dict_i = model.validation_step(eval_data)
                loss.append(loss_i.numpy())
                for key, error_i in error_dict_i.items():
                    error_i = error_i.numpy()
                    if np.isfinite(error_i):
                        error_dict[key].append(error_i)

                if self.enable_progress_bar:
                    self.eval_pbar.update(1)
                    descriptions = [self.callback_pbar('val/loss', loss[i])]
                    self.eval_pbar.set_postfix_str(descriptions)
                    self.eval_pbar.refresh() 
            self.eval_current_index = [0 for i in range(len(eval_dataloader))]
            if self.enable_progress_bar:
                self.eval_pbar.clear()
                self.eval_pbar.close()
        else:
            loss, error_dict = model.validation_step(eval_dataloader)

        loss = np.mean(loss)
        self.set_callback_metrics('val/loss', loss)
        
        if self.enable_progress_bar:
            descriptions = [self.callback_pbar('val/loss', loss)]
            for error_name in model.val_solution_names:
                error_dict[error_name] = np.mean(error_dict[error_name])
                descriptions.append(self.callback_pbar(f'val/error_{error_name}',
                                                       error_dict[error_name]))
            
            # Join all descriptions into a single string and set it
            full_description = ', '.join(descriptions)
            self.pbar.set_postfix_str(full_description)
            self.pbar.refresh() 
        return loss, error_dict

    def pred_loop(self, model, dataloader):
        """
        Prediction loop. 

        :param model: The PINNModule.
        :param dataloader: The prediction data.
        """
        preds = {}
        
        if self.batch_size is not None:
            
            iter = (self.pred_dataset_size[0]//self.batch_size) + 1
            
            if self.enable_progress_bar:
                self.pred_pbar = self.initalize_tqdm(iter, mode='prediction', leave=False)
            
            for i in range(iter):
                pred_data, self.pred_current_index = self.next_batch(dataloader,
                                                                     self.pred_current_index,
                                                                     self.pred_dataset_size,
                                                                     shuffle=False)
                preds_i = model.predict_step(pred_data)

                for key, value in preds_i.items():
                    if key in preds.keys():
                        preds[key] = np.concatenate(preds[key], preds_i[key].numpy(), 1)
                    else:
                        preds[key] = preds_i[key].numpy()
                        
                if self.enable_progress_bar:
                    self.pred_pbar.update(1)
                    self.pred_pbar.refresh() 
            
            self.pred_current_index = [0 for i in range(len(dataloader))]
            
            if self.enable_progress_bar:
                self.pred_pbar.clear()
                self.pred_pbar.close()
        else:
            preds = model.predict_step(dataloader)
            for sol_key, pred in preds.items():
                preds[sol_key] = pred.numpy()
                
        return preds

    def validate(self, model, datamodule):
        """
        Validate the model using the provided data module.

        :param model: The PINNModule.
        :param datamodule: The data module providing the validation data.
        :return: Tuple of loss and error dictionary from validation.
        """
        datamodule.setup('val')
        val_dataloader = datamodule.val_dataloader()
        model = self.prepare_model(model, datamodule)
        
        loss, error_dict = self.eval_loop(model, val_dataloader)

        for key, error in error_dict.items():   
            self.set_callback_metrics(f'val/error_{key}', error)
        
        return loss, error_dict

    def predict(self, model, datamodule):
        """
        Generate predictions using the model and data module.

        :param model: The PINNModule.
        :param datamodule: The data module providing prediction data.
        :return: Predictions made by the model.
        """
        
        datamodule.setup('pred')
        pred_dataloader = datamodule.predict_dataloader()
        model = self.prepare_model(model, datamodule)

        if datamodule.batch_size is not None:
            self.pred_current_index, self.pred_dataset_size = self.initalize_batch_tracker(pred_dataloader, shuffle=False)

        
        preds = self.pred_loop(model, pred_dataloader)
        
        return preds

    def test(self, model, datamodule):
        """
        Test the model using the provided data module.

        :param model: The PINNModule.
        :param datamodule: The data module providing the test data.
        :return: Tuple of loss and error dictionary from testing.
        """
        
        datamodule.setup('test')
        test_dataloader = datamodule.test_dataloader()
        model = self.prepare_model(model, datamodule)

        if datamodule.batch_size is not None:
            self.eval_current_index, self.eval_dataset_size = self.initalize_batch_tracker(test_dataloader, shuffle=False)

        loss, error_dict = self.eval_loop(model, test_dataloader, mode='test')

        for key, error in error_dict.items():   
            self.set_callback_metrics(f'test/error_{error}', error)
        
        return loss, error_dict


from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import os
import time
import hydra
import rootutils
import numpy as np
import tensorflow as tf

from omegaconf import DictConfig, OmegaConf

rootutils.setup_root(__file__, dotenv=True, pythonpath=True)

from pinnstf2.trainer import Trainer
from pinnstf2 import utils
from pinnstf2.models import PINNModule
from pinnstf2.data import (
    Interval,
    Mesh,
    PointCloud,
    Rectangle,
    RectangularPrism,
    TimeDomain,
    PINNDataModule,
)

log = utils.get_pylogger(__name__)

OmegaConf.register_new_resolver("eval", eval)

@utils.task_wrapper
def train(
    cfg: DictConfig, read_data_fn: Callable, pde_fn: Callable, output_fn: Callable = None
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Trains the model. Can additionally evaluate on a testset, using best weights obtained during
    training.

    This method is wrapped in optional @task_wrapper decorator, that controls the behavior during
    failure. Useful for multiruns, saving info about the crash, etc.

    :param cfg: A DictConfig configuration composed by Hydra.
    :return: A tuple with metrics and dict with all instantiated objects.
    """

    cfg = utils.set_mode(cfg)

    
    # set seed for random number generators in pytorch, numpy and python.random
    if cfg.get("seed"):
        np.random.seed(cfg.seed)
        tf.random.set_seed(cfg.seed)

    if cfg.get("time_domain"):
        log.info(f"Instantiating time domain <{cfg.time_domain._target_}>")
        td: TimeDomain = hydra.utils.instantiate(cfg.time_domain)

    if cfg.get("spatial_domain"):
        log.info(f"Instantiating spatial domain <{cfg.spatial_domain._target_}>")
        sd: Union[Interval, Rectangle, RectangularPrism] = hydra.utils.instantiate(
            cfg.spatial_domain
        )

    log.info(f"Instantiating mesh <{cfg.mesh._target_}>")
    if cfg.mesh._target_ == "pinnstf2.data.Mesh":
        mesh: Mesh = hydra.utils.instantiate(
            cfg.mesh, time_domain=td, spatial_domain=sd, read_data_fn=read_data_fn
        )
    elif cfg.mesh._target_ == "pinnstf2.data.PointCloud":
        mesh: PointCloud = hydra.utils.instantiate(cfg.mesh, read_data_fn=read_data_fn)
    else:
        raise "Mesh should be defined in config file."

    train_datasets = []
    for i, (dataset_dic) in enumerate(cfg.train_datasets):
        for key, dataset in dataset_dic.items():
            log.info(f"Instantiating training dataset number {i+1}: <{dataset._target_}>")
            train_datasets.append(hydra.utils.instantiate(dataset)(mesh=mesh, dtype = cfg.dtype))

    val_dataset = None
    if cfg.get("val_dataset"):
        for i, dataset_dic in enumerate(cfg.val_dataset):
            for key, dataset in dataset_dic.items():
                log.info(f"Instantiating validation dataset number {i+1}: <{dataset._target_}>")
                val_dataset = hydra.utils.instantiate(dataset)(mesh=mesh, dtype = cfg.dtype)

    test_dataset = None
    if cfg.get("test_dataset"):
        for i, dataset_dic in enumerate(cfg.test_dataset):
            for key, dataset in dataset_dic.items():
                log.info(f"Instantiating test dataset number {i+1}: <{dataset._target_}>")
                test_dataset = hydra.utils.instantiate(dataset)(mesh=mesh, dtype = cfg.dtype)

    pred_dataset = None
    if cfg.get("pred_dataset"):
        for i, dataset_dic in enumerate(cfg.pred_dataset):
            for key, dataset in dataset_dic.items():
                log.info(f"Instantiating prediction dataset number {i+1}: <{dataset._target_}>")
                pred_dataset = hydra.utils.instantiate(dataset)(mesh=mesh, dtype = cfg.dtype)

    log.info(f"Instantiating datamodule <{cfg.data._target_}>")
    datamodule: PINNDataModule = hydra.utils.instantiate(
        cfg.data,
        train_datasets=train_datasets,
        val_dataset=val_dataset,
        test_dataset=test_dataset,
        pred_dataset=pred_dataset,
        batch_size=cfg.get("batch_size"),
    )

    if cfg.net._target_ == "pinnstf2.models.FCN":
        log.info(f"Instantiating neural net <{cfg.net._target_}>")
        net = hydra.utils.instantiate(cfg.net)(lb=mesh.lb, ub=mesh.ub, dtype = cfg.dtype)
    elif cfg.net._target_ == "pinnstf2.models.NetHFM":
        # TODO
        log.info(f"Instantiating neural net <{cfg.net._target_}>")
        net = hydra.utils.instantiate(cfg.net)(
            mean=train_datasets[0].mean, std=train_datasets[0].std
        )

    log.info(f"Instantiating model <{cfg.model._target_}>")

    model: PINNModule = hydra.utils.instantiate(cfg.model)(
        net=net, pde_fn=pde_fn, output_fn=output_fn, dtype = cfg.dtype
    )

    log.info(f"Instantiating trainer <{cfg.trainer._target_}>")
    trainer: Trainer = hydra.utils.instantiate(cfg.trainer)

    object_dict = {
        "cfg": cfg,
        "datamodule": datamodule,
        "model": model,
        "trainer": trainer,
    }

    if cfg.get("train"):
        log.info("Starting training!")
        start_time = time.time()
        try:
            trainer.fit(model=model, datamodule=datamodule)
        except KeyboardInterrupt:
            print("Training Stopped.")
        log.info(f"Elapsed time: {time.time() - start_time}")
    log.info(f"MEAN time: {np.median(trainer.time_list[-5:])}  -   {np.median(trainer.time_list[-100:-10])} -  {np.median(trainer.time_list)}")

    

    if cfg.get("val"):
        log.info("Starting validation!")
        trainer.validate(model=model, datamodule=datamodule)

    train_metrics = trainer.callback_metrics

    if cfg.get("test"):
        log.info("Starting testing!")
        ckpt_path = None  # trainer.checkpoint_callback.best_model_path
        trainer = hydra.utils.instantiate(cfg.trainer)
        trainer.test(model=model, datamodule=datamodule, ckpt_path=ckpt_path)

    test_metrics = trainer.callback_metrics
    
    if cfg.get("plotting"):
        try:
            log.info("Plotting the results")
            preds_dict = trainer.predict(
                model=model, datamodule=datamodule
            )
            hydra.utils.instantiate(
                cfg.plotting,
                mesh=mesh,
                preds=preds_dict,
                train_datasets=train_datasets,
                val_dataset=val_dataset,
                file_name=cfg.paths.output_dir,
            )()
        except KeyboardInterrupt:
            print("Plotting Stopped.")
    # merge train and test metrics
    metric_dict = {**train_metrics, **test_metrics}

    return metric_dict, object_dict


@hydra.main(version_base="1.3", config_path="conf", config_name="train.yaml")
def main(cfg: DictConfig) -> Optional[float]:
    """Main entry point for training.

    :param cfg: DictConfig configuration composed by Hydra.
    :return: Optional[float] with optimized metric value.
    """
    # apply extra utilities
    # (e.g. ask for tags if none are provided in cfg, print cfg tree, etc.)
    utils.extras(cfg)

    # train the model
    metric_dict, _ = train(cfg)

    # safely retrieve metric value for hydra-based hyperparameter optimization
    metric_value = utils.get_metric_value(
        metric_dict=metric_dict, metric_name=cfg.get("optimized_metric")
    )

    # return optimized metric
    return metric_value


if __name__ == "__main__":
    main()
